#!/usr/bin/perl

use strict;
use Net::DNS;
use CGI;

###############################################################################
# dns-propagation-tracker.pl - queries various DNS servers for the given 
# hostname, and shows the answers from each.
# 
# Useful to see the propagation of a DNS change.
#
# Copyright (C) 2006-2008, David Precious <davidp@preshweb.co.uk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# $Id: dns-propagation-tracker.pl 400 2008-08-06 16:01:54Z davidp $
###############################################################################
# Configuration:

# how many servers to query:
my $num = 10;

# File containing DNS servers we can use; if you place the file in a different
# directory to this script, you must update this:
my $server_list_file = 'dns-resolvers-list';

# This tool uses the jQuery Javascript library.  Please change the URL below
# to a local copy of the jQuery library - see www.jquery.com for a copy.
my $jquery_url = 'http://www.preshweb.co.uk/jquery.js';

# End of configuration.
###############################################################################


$| = 1; # don't buffer output... start outputting results as they
        # come so the user can see we're doing something

my $cgi = new CGI;

my $domain = $cgi->param('domain');
my $type   = $cgi->param('type') || 'A';

if ($domain =~ /([0-9]{1,3}\.){3}[0-9]{1,3}/) {
    # it's an IP, so we probably want a PTR lookup :)
    $type = 'PTR';
}

get_domain() unless $domain;


my @servers;
my $serverstxt;

if ($cgi->param('serversmode') eq 'specified') {
    # we're being given a set of servers to query:
    @servers = split /\n/, $cgi->param('servers');
    s/(\r|\n)*$// for @servers;
    @servers = map { { ip => $_, host => $_, note => '' } } @servers;
          
    $serverstxt = "provided set of " . scalar @servers . " servers";

} elsif ($cgi->param('serversmode') eq 'nsdomain')  {

    my $resolver = Net::DNS::Resolver->new;

    my $answer = $resolver->query($cgi->param('nsdomain'), 'NS');
    
    for my $rr (grep { $_->{type} eq 'NS' } $answer->answer) {
        push @servers, { host => $rr->nsdname };
    }

    $serverstxt = "nameservers for domain " . $cgi->param('nsdomain')
	. ' (' . join(',', map { $_->{host} } @servers) . ')';

} else {
    
    # read in server list:
    open(SERVERLST, $server_list_file)
        || die "Failed to open server list\n";
    
    my @available_servers;
    
    LINE:
    for my $line (<SERVERLST>) {
    
        next LINE if $line =~ /^(#|$)/;
        
        my %data;
        
        @data{qw(ip host note)} = split /\s+/, $line, 3;
        push @available_servers, {%data};    
    }
    close SERVERLST;
    
    $serverstxt = "a random set of $num servers out of "
        . scalar @available_servers . " available servers";
    
    @servers = ((sort { (-1,1)[rand 2] } @available_servers)[0 .. $num]);
}


my %data;

SERVER:

print page_start("Propagation results for $domain");


if (!@servers) {
    print "<p>FAILED to identify nameservers for " . $cgi->param('nsdomain') . "</p>";
    print page_end();
    exit;
}



print qq[
<p>Querying $serverstxt.  Please be patient, sometimes this can take
a while if some servers don't respond.</p>
<table border="1" cellpadding="1">
<tr>
    <th>DNS server</th>
    <th>Class</th>
    <th>Host</th>
    <th>Value</th>
    <th>TTL</th>
</tr>
];

# we'll keep a list of answers we see, so we can report whether all servers
# agreed.
my %unique_answers;

SERVER:
for my $server (@servers) {

    my $resolver = Net::DNS::Resolver->new(
	nameservers => [ $server->{host} ],
	recurse     => 1,
	debug       => 0,
	retry       => 1,
    );
    
    $resolver->tcp_timeout(4);    
    
    my $result = $resolver->query($domain, $type);
    
    if (!$result) {
        # we didn't get a result from this server:
        show_result($server, [] , 'No result');
        next SERVER;
    }
    
    my @answers = $result->answer;
       
    my @answer_list; # build up list of answers to be displayed
    my @t;
    for my $answer (@answers) {
        my %answerdata;
        @answerdata{qw(host ttl type class val)} =
            split /\s+/, $answer->string, 5;
        
        push @answer_list, {%answerdata};
        push @t, $answerdata{host};
        
    }
    $unique_answers{ join ',', sort @t }++;
    show_result($server, \@answer_list);

}

print "\n</table>\n";

my $num_diff_answers = scalar keys %unique_answers;
if ($num_diff_answers == 1) {
    print <<ALLAGREE;
<p style="color:green; font-weight:bold;">
All responding servers agreed on the same answer
</p>
ALLAGREE
} elsif ($num_diff_answers == 0) {
    print "<p>No answers were received.</p>";
} else {
    print <<DIFFANSWERS;
<p style="color:red; font-weight:bold;">
Not all servers agreed, $num_diff_answers different answers were received.</p>
DIFFANSWERS

}

print q[
<p>
End of results.
</p>
];

print page_end();
exit;


sub show_result {

    my ($server, $answer_list, $error) = @_;
        
    my $rowspan = ($error)? 1 : @$answer_list;
    
    printf qq[<tr><td rowspan="%d">%s<br /><small>%s</small></td>], 
        $rowspan, @$server{ qw(host note) };
    
    if ($error) {
        print qq[<td colspan="4">$error</td></tr>];
    } else {
              
        my $ans_num = 0;
        for my $answer (@$answer_list) {
            # $answer is a hashref
            print "<tr>" if $ans_num > 0;
            print "<td>";
            print join '</td><td>',  @$answer{ qw(class host val ttl) };
            print "</td></tr>";
                
        }
        
        print "</tr>";
        
        
    }
    
}





sub get_domain() {

    print page_start('DNS Propagation Checker');

print q[
<p>
This is a simple tool to query a long list of DNS servers, and check that they
all return the correct info for the hostname you're interested in.
</p>

<p>
If all (or most) of the servers checked by this tool indicate the same desired
result, then you can be pretty sure that your DNS change has propagated across
most of the net.
</p>

<p>
I need to update the DNS servers list, and trim out some old servers which seem
to no longer work.  I'll do that soon.  If you have any servers you'd like to
see on the list, mail me the details.
</p>

<p>
Coming soon: ability to check against every server (probably returning results
by email, as it may take a while, and I don't want to send too many DNS results
out at once, that's opening up a potential DoS attack).
</p>


<p>
Enter the host/IP to look up and the type of record.  If you enter an IP,
the type selected will be ignored, and a PTR lookup will be automatically 
performed.
</p>

<p>
If you'd like to query a specific set of servers rather than a random selection 
from my list, select "specified servers".
</p>

<form method="get" name="lookupform">
<div style="padding: 10px; background-color: #ada;">
Hostname: <input type="text" name="domain" />
Type: <select name="type">
<option value="A">A (or PTR) (normal address lookups)</option>
<option value="MX">MX (incoming mail servers)</option>
<option value="NS">NS (nameservers)</option>
<option value="TXT">TXT (text records, eg SPF)</option>
</select>
<br />
<label><input type="radio" name="serversmode" value="random"
    onchange="toggle_servers_visibility()">
 Query a random set of servers</label>
<label><input type="radio" name="serversmode" value="specified"
    onchange="toggle_servers_visibility()">
 Query a set of servers I specify</label>
<label><input type="radio" name="serversmode" value="nsdomain"
    onchange="toggle_servers_visibility()">
 Query all the nameservers for a given domain
<div id="servers">
Servers to query:<br />
<textarea name="servers" rows="6" cols="50"></textarea>
</div>
<div id="nsdomain">
Domain to get nameservers from:<br />
<input type="text" name="nsdomain" />
</div>

<br />
<input type="submit" value="Check now" />
</div>
</form>

<p>
The code to this tool is Open Source, released under the terms of the 
<a href="http://www.gnu.org/licenses/agpl.txt">
GNU Affero General Public Licence</a>, version 3.
</p>

<p>
You can download the source:
<a href="http://www.preshweb.co.uk/downloads/dns-propagation-tracker-0.02.tar.gz">
dns-propagation-tracker-0.02.tar.gz</a>.  If you use it, I'd greatly appreciate
a quick email to <a href="mailto:davidp@preshweb.co.uk">davidp@preshweb.co.uk</a>
to let me know where you're using it and what for, but this is of course not 
required.
</p>
];
print qq[<script src="$jquery_url"></script>];
print q[

<script type="text/javascript">

function toggle_servers_visibility() {
    var serversmode = radio_group_get_value(document.lookupform.serversmode);
    if (serversmode == 'specified') {
        $('#servers').show('slow');
        $('#nsdomain').hide('slow');
    } else if (serversmode == 'nsdomain') {
        $('#nsdomain').show('slow');
        $('#servers').hide('slow');
    } else {
        $('#nsdomain').hide('slow');
        $('#servers').hide('slow');
    }
}

toggle_servers_visibility();



function radio_group_get_value(group) {
// gets the value of the currently-checked radio button in the
// radio group 'group', or returns false if none is ticked (or the
// radio group didn't exist).

// check the group exists:
if (!group) { return false; }

// if there's more than one radio button in the group it will
// be treated as an array, test for this with the length property:
if (group.length) 
        { // more than one - loop through
        for (x=0; x<group.length;x++)
                {
                if (group[x].checked) { return group[x].value; }
                }
        } else {
        // there's only one radio box - see if it's ticked
        if (group.checked) { return group.value; }
        } // end of if (group.length)
return false; // didn't find it

} // end of function radio_group_get_value()

</script>
];
print page_end();
exit;
}


sub page_start() {

    my $title = shift;

qq[Content-Type: text/html

<html>
<head>
<title>$title</title>
</head>

<body>
<h1>$title</h1>
];
}

sub page_end {
    q[
DNS Propagation Checker by David Precious 
(<a href="mailto:davidp@preshweb.co.uk">davidp@preshweb.co.uk</a>)
<pre>
$Id: dns-propagation-tracker.pl 400 2008-08-06 16:01:54Z davidp $
</pre>


</body>
</html>
];
}
